package chocopy.pa3;

import java.nio.charset.StandardCharsets;
import java.util.*;

import chocopy.common.analysis.SymbolTable;
import chocopy.common.analysis.AbstractNodeAnalyzer;
import chocopy.common.analysis.types.SymbolType;
import chocopy.common.analysis.types.ValueType;
import chocopy.common.astnodes.*;
import chocopy.common.codegen.*;

import static chocopy.common.codegen.RiscVBackend.Register.*;

/**
 * This is where the main implementation of PA3 will live.
 *
 * A large part of the functionality has already been implemented
 * in the base class, CodeGenBase. Make sure to read through that
 * class, since you will want to use many of its fields
 * and utility methods in this class when emitting code.
 *
 * Also read the PDF spec for details on what the base class does and
 * what APIs it exposes for its sub-class (this one). Of particular
 * importance is knowing what all the SymbolInfo classes contain.
 */
public class CodeGenImpl extends CodeGenBase {
    private class BetterRsicVBackend {
        private final RiscVBackend backend;

        BetterRsicVBackend(RiscVBackend _backend) {
            backend = _backend;
        }

        public void emitNoop(String comment) {
            if (comment != null) {
                backend.emitMV(ZERO, ZERO, "Noop: " + comment);
            }
        }

        public void emitPush(RiscVBackend.Register reg, String comment) {
            if(comment == null) comment = "";
            comment = "PUSH: " + comment;
            backend.emitADDI(SP, SP, -1 * backend.getWordSize(), comment);
            backend.emitSW(reg, SP, 0, comment);
        }

        public void emitPop(RiscVBackend.Register reg, String comment) {
            if(comment == null) comment = "";
            comment = "POP: " + comment;
            if (reg != null)
                backend.emitLW(reg, SP, 0, comment);
            backend.emitADDI(SP, SP, backend.getWordSize(), comment);
        }

        public void emitCall(Label calledLabel, String comment) {
            // Arguments should be already pushed to stack.
            backend.emitJAL(calledLabel, comment);
        }

        public void emitFunctionBegin(String funcName, String comment) {
            emitNoop(comment);
            emitPush(RA, "backup my return address.");
            emitPush(FP, "backup parent fp.");
            backend.emitADDI(FP, SP, 2 * backend.getWordSize(), "Change FP to apply new frame.");

            backend.emitJ(new Label("$" + funcName + "$saveReg"), "Jump to save callee-saved registers");
            backend.emitLocalLabel(new Label("$" + funcName + "$saveRegContinue"), "Begin real function codes:");
        }

        public void emitFunctionEnd(String funcName, Void regMgr, String comment) {
            // Return value should be in A0
            List<RiscVBackend.Register> savedRegs = new ArrayList<RiscVBackend.Register>() {}; // regMgr.registerToSaveAndRestoreInFunc;

            // Function return point:
            backend.emitLocalLabel(new Label("$" + funcName + "$exitPoint"), "Function return point");

            emitNoop(comment);
            backend.emitADDI(SP, FP, -(2 + savedRegs.size()) * backend.getWordSize(), "Revert all local variables on this dying frame.");
            for (int i = savedRegs.size(); i-- > 0; ) {
                emitPop(savedRegs.get(i), "restore callee-saved reg.");
            }
            emitPop(FP, "restore parent FP.");
            emitPop(RA, "restore my return address.");
            backend.emitJR(RA, "Real return!");

            backend.emitLocalLabel(new Label("$" + funcName + "$saveReg"), "Save callee-saved registers before use.");
            for (int i = 0; i < savedRegs.size(); ++i) {
                emitPush(savedRegs.get(i), "save callee-saved reg.");
            }
            backend.emitJ(new Label("$" + funcName + "$saveRegContinue"), "go back to execute real function code.");
        }

        public void emitFunctionReturn(String funcName, String comment) {
            backend.emitJ(new Label("$" + funcName + "$exitPoint"), comment);
        }

        public void emitWrapIntValToIntObject(RiscVBackend.Register tmpReg, SymbolType typeIntOrBool) {
            emitPop(tmpReg, "INT VAL TO INT OBJ: POP");
            emitAllocPrototype(typeIntOrBool, tmpReg);
            emitPush(A0, "INT VAL TO INT OBJ: PUSH");
        }

        public void emitPushStrVal(RiscVBackend.Register tmpReg, String val, String comment) {
            emitNoop(comment);
            byte[] bytes = val.getBytes(StandardCharsets.US_ASCII);
            int dataLen = bytes.length / backend.getWordSize() + 1; // Equals to ceil((len+1) DIVIDE WORD_SIZE)

            for (int cter = 0; cter < dataLen; ++cter) {
                int myRangeBegin = (dataLen - cter - 1) * backend.getWordSize();
                int myRangeEnd = Math.min(myRangeBegin + backend.getWordSize(), bytes.length);
                int curr = 0;
                assert backend.getWordSize() == 4; // curr should be 4 byte = int
                String debug_pushed_str = "";
                for (int shift = 0; shift < myRangeEnd - myRangeBegin; ++shift) {
                    curr += bytes[myRangeBegin + shift] << (shift * 8);
                    debug_pushed_str += Integer.toString(bytes[myRangeBegin + shift]) + " ";
                }

                backend.emitLI(tmpReg, curr, "STR VAL - " + debug_pushed_str);
                emitPush(tmpReg, "STR VAL - PUSH A BLOCK");
            }

            backend.emitLI(tmpReg, bytes.length, "STR ATTR - __len__");
            emitPush(tmpReg, null);
            backend.emitLA(tmpReg, strClass.getDispatchTableLabel(), "OBJECT HEAD - DISPATCH TBL");
            emitPush(tmpReg, null);
            backend.emitLI(tmpReg, 3 + dataLen + 1, "OBJECT HEAD - SIZE");
            emitPush(tmpReg, null);
            backend.emitLI(tmpReg, 3, "OBJECT HEAD - TYPE = STR");
            emitPush(tmpReg, null);
            backend.emitMV(tmpReg, SP, "Return STR OBJ address.");
        }

        private Void emitAllocPrototype(SymbolType typ, RiscVBackend.Register objReg) {
            // Convention: if typ is INT_TYPE or BOOL_TYPE then objReg contains the literal value, not a pointer.
            // @returns A0
            if (typ.equals(SymbolType.INT_TYPE)) {
                backend.emitLA(A0, intClass.getPrototypeLabel(), "Load prototype address");
                // save our objReg on stack before `alloc`
                betterBackend.emitPush(objReg, "Store the literal value");
                backend.emitJAL(objectAllocLabel, "Allocate int");
                betterBackend.emitPop(objReg, "Load literal value again");
                backend.emitSW(objReg, A0, 3 * backend.getWordSize(), "Set __int__ value");
            } else if (typ.equals(SymbolType.STR_TYPE)) {
                // objReg already has a ptr in this case. just move to A0.
                backend.emitMV(A0, objReg, "move string ptr to A0");
            } else if (typ.equals(SymbolType.BOOL_TYPE)){
                backend.emitLA(A0, constants.getBoolConstant(false), "Load FALSE bool");
                backend.emitSLLI(objReg, objReg, 4, "Compute offset to correct object");
                backend.emitADD(A0, A0, objReg, "Add offset to curr pointer");
            }
            return null;
        }

    }


    /** A code generator emitting instructions to BACKEND. */
    public CodeGenImpl(RiscVBackend backend) {
        super(backend);
        betterBackend = new BetterRsicVBackend(backend);
    }

    private BetterRsicVBackend betterBackend;

    /** Operation on None. */
    private final Label errorNone = new Label("error.None");
    /** Division by zero. */
    private final Label errorDiv = new Label("error.Div");
    /** Index out of bounds. */
    private final Label errorOob = new Label("error.OOB");

    private Map<FuncInfo, StmtAnalyzer> mapFuncInfoToStmtAnalyzer = new HashMap<>();

    /**
     * Emits the top level of the program.
     *
     * This method is invoked exactly once, and is surrounded
     * by some boilerplate code that: (1) initializes the heap
     * before the top-level begins and (2) exits after the top-level
     * ends.
     *
     * You only need to generate code for statements.
     *
     * @param statements top level statements
     */
    protected void emitTopLevel(List<Stmt> statements) {
        StmtAnalyzer stmtAnalyzer = new StmtAnalyzer(null);
        backend.emitADDI(SP, SP, -2 * backend.getWordSize(),
                         "Saved FP and saved RA (unused at top level).");
        backend.emitSW(ZERO, SP, 0, "Top saved FP is 0.");
        backend.emitSW(ZERO, SP, 4, "Top saved RA is 0.");
        backend.emitADDI(FP, SP, 2 * backend.getWordSize(),
                         "Set FP to previous SP.");
        for (Stmt stmt : statements) {
            stmt.dispatch(stmtAnalyzer);
        }
        backend.emitLI(A0, EXIT_ECALL, "Code for ecall: exit");
        backend.emitEcall(null);
    }

    /**
     * Emits the code for a function described by FUNCINFO.
     *
     * This method is invoked once per function and method definition.
     * At the code generation stage, nested functions are emitted as
     * separate functions of their own. So if function `bar` is nested within
     * function `foo`, you only emit `foo`'s code for `foo` and only emit
     * `bar`'s code for `bar`.
     */
    protected void emitUserDefinedFunction(FuncInfo funcInfo) {
        backend.emitGlobalLabel(funcInfo.getCodeLabel());
        StmtAnalyzer stmtAnalyzer;
        if(mapFuncInfoToStmtAnalyzer.containsKey(funcInfo)) {
            // LocalTable(analyzer) already inited by this function's child.
            stmtAnalyzer = mapFuncInfoToStmtAnalyzer.get(funcInfo);
        }
        else {
            stmtAnalyzer = new StmtAnalyzer(funcInfo);
        }

        betterBackend.emitFunctionBegin(funcInfo.getFuncName(), null);
        stmtAnalyzer.pushLocalVars(funcInfo.getLocals()); // regMgr should NEVER be used, because of this statement.

        // Emit code for function stmts
        for (Stmt stmt : funcInfo.getStatements()) {
            stmt.dispatch(stmtAnalyzer);
        }

        betterBackend.emitFunctionEnd(funcInfo.getFuncName(), null, null);
    }

    /** An analyzer that encapsulates code generation for statments. */
    private class StmtAnalyzer extends AbstractNodeAnalyzer<Void> {
        /*
         * The symbol table has all the info you need to determine
         * what a given identifier 'x' in the current scope is. You can
         * use it as follows:
         *   SymbolInfo x = sym.get("x");
         *
         * A SymbolInfo can be one the following:
         * - ClassInfo: a descriptor for classes
         * - FuncInfo: a descriptor for functions/methods
         * - AttrInfo: a descriptor for attributes
         * - GlobalVarInfo: a descriptor for global variables
         * - StackVarInfo: a descriptor for variables allocated on the stack,
         *      such as locals and parameters
         *
         * Since the input program is assumed to be semantically
         * valid and well-typed at this stage, you can always assume that
         * the symbol table contains valid information. For example, in
         * an expression `foo()` you KNOW that sym.get("foo") will either be
         * a FuncInfo or ClassInfo, but not any of the other infos
         * and never null.
         *
         * The symbol table in funcInfo has already been populated in
         * the base class: CodeGenBase. You do not need to add anything to
         * the symbol table. Simply query it with an identifier name to
         * get a descriptor for a function, class, variable, etc.
         *
         * The symbol table also maps nonlocal and global vars, so you
         * only need to lookup one symbol table and it will fetch the
         * appropriate info for the var that is currently in scope.
         */

        /** Symbol table for my statements. */
        private SymbolTable<SymbolInfo> sym;

        /** Label of code that exits from procedure. */
        protected Label epilogue;

        /** The descriptor for the current function, or null at the top
         *  level. */
        private FuncInfo funcInfo;

        private int _r_local_label_counter = 427;

        // Extend StackVarInfo to add more info.
        private Map<String, Integer> localMap = new HashMap<>(); // Map from StackVarInfo to offset of FP.
        private void initArgsMap(List<String> args) {
            for(int i = 0; i < args.size(); ++i) {
                int offset = (args.size() - i - 1) * backend.getWordSize();
                localMap.put(args.get(i), offset);
            }
        }
        private void initLocalVars(List<StackVarInfo> locals, int firstLocalOffset) {
            for(int i = 0; i < locals.size(); ++i) {
                StackVarInfo theLocal = locals.get(i);
                int offset = firstLocalOffset - i * backend.getWordSize();
                localMap.put(theLocal.getVarName(), offset);
            }
        }
        public void pushLocalVars(List<StackVarInfo> locals) {
            // This function generates some asm code, so it should only be called at RIGHT time.
            for(int i = 0; i < locals.size(); ++i) {
                StackVarInfo theLocal = locals.get(i);
                theLocal.getInitialValue().dispatch(this);
                // The literal initial value is pushed to stack. That's the local!
                // No need to push it again.
            }
        }
        private void emitLoadNestedParentVar(StackVarInfo varInfo, RiscVBackend.Register tmpReg, RiscVBackend.Register valReg_to_assign) {
            // To assign: set valReg properly.
            // To read the var: set valReg to null. Returns valReg.
            FuncInfo hisOwner = varInfo.getFuncInfo();

            if(!mapFuncInfoToStmtAnalyzer.containsKey(hisOwner)) {
                new StmtAnalyzer(hisOwner);
            }

            //! assert: hisOwner is parent_of(this)
            FuncInfo curr = this.funcInfo;
            int counter = 0;
            while(curr != hisOwner && curr != null) {
                curr = curr.getParentFuncInfo();
                ++counter;
            }
            assert curr != null;
            assert counter == funcInfo.getDepth() - hisOwner.getDepth();
            //! assert end

            int depthDiff = funcInfo.getDepth() - hisOwner.getDepth();
            // If depthDiff is 0, then the variable is in current frame.
            backend.emitMV(tmpReg, FP, "LOAD NESTED VAR: Load current FP.");
            for(int i = 0; i < depthDiff; ++i) {
                backend.emitLW(tmpReg, tmpReg, -2 * backend.getWordSize(), "LOAD NESTED VAR: Load parent FP.");
            }
            int offsetInParent = mapFuncInfoToStmtAnalyzer.get(hisOwner).localMap.get(varInfo.getVarName());
            if(valReg_to_assign == null) {
                // Read the value to tmpReg
                backend.emitLW(tmpReg, tmpReg, offsetInParent, "LOAD NESTED VAR: Read val: Set the result!");
            }
            else {
                // Write to value:
                backend.emitSW(valReg_to_assign, tmpReg, offsetInParent, "LOAD NESTED VAR: Write val: Set the val!");
            }
        }

        /** An analyzer for the function described by FUNCINFO0, which is null
         *  for the top level. */
        StmtAnalyzer(FuncInfo funcInfo0) {
            funcInfo = funcInfo0;
            if (funcInfo == null) {
                sym = globalSymbols;
            } else {
                sym = funcInfo.getSymbolTable();
            }

            if(funcInfo != null) {
                // The init function below won't generate any asm code, so the stmtAnalyzer can be constructed at any time.
                // because the constructor doesn't generate code! It won't affect anything.
                initArgsMap(funcInfo.getParams());
                initLocalVars(funcInfo.getLocals(), -3 * backend.getWordSize());
                mapFuncInfoToStmtAnalyzer.put(funcInfo, this);
            }
        }

        @Override
        public Void analyze(ExprStmt exprStmt) {
            exprStmt.expr.dispatch(this);
            return null;
        }

        @Override
        public Void analyze(AssignStmt node) {
            node.value.dispatch(this);
            RiscVBackend.Register tmpReg = /*regMgr.borrowOneTmp()*/ T6;
            RiscVBackend.Register tmpReg2 = /*regMgr.borrowOneTmp()*/ T5;
            betterBackend.emitPop(tmpReg, "ASSIGN: Get the value to assign: Pop the expr result from stack");

            SymbolType symbolType = null;
            for(Expr expr : node.targets) {
                if(! (expr instanceof Identifier)) {
                    throw new RuntimeException("assignment target should be Identifier.");
                }
                SymbolInfo symbolInfo = sym.get(((Identifier) expr).name);
                Label label = null;
                if(symbolType == null) {
                    // first run:
                    if (symbolInfo instanceof GlobalVarInfo) {
                        GlobalVarInfo globalVarInfo = (GlobalVarInfo) symbolInfo;
                        symbolType = globalVarInfo.getVarType();
                    } else if (symbolInfo instanceof StackVarInfo) {
                        StackVarInfo stackVarInfo = (StackVarInfo) symbolInfo;
                        symbolType = stackVarInfo.getVarType();
                    }
                    if (symbolType.equals(SymbolType.INT_TYPE) || symbolType.equals(SymbolType.BOOL_TYPE)) {
                        // first run:
                        /* All int and bool variable is not wrapped.
                        backend.emitLW(tmpReg, tmpReg, 3 * backend.getWordSize(), "ASSIGN: Load the value of INT or BOOL");
                        */
                    }
                }

                // do assign
                if (symbolInfo instanceof GlobalVarInfo) {
                    GlobalVarInfo globalVarInfo = (GlobalVarInfo) symbolInfo;
                    backend.emitSW(tmpReg, globalVarInfo.getLabel(), tmpReg2, "ASSIGN: Assign to global var");
                } else if (symbolInfo instanceof StackVarInfo) {
                    StackVarInfo stackVarInfo = (StackVarInfo) symbolInfo;
                    this.emitLoadNestedParentVar(stackVarInfo, tmpReg2, tmpReg); // Also work for local variable or param.
                }
            }
            return null;
        }

        private String getLablePrefix(String name) {
            String labelPrefix = "";
            if(funcInfo != null) {
                labelPrefix += "$" + funcInfo.getFuncName();
            }
            labelPrefix += "$" + name + _r_local_label_counter;
            ++_r_local_label_counter;
            return labelPrefix;
        }

        @Override
        public Void analyze(IfStmt node) {
            node.condition.dispatch(this);
            RiscVBackend.Register tmpReg = /*regMgr.borrowTmp()*/ T6;
            betterBackend.emitPop(tmpReg, "IfStmt: Get condition");

            String labelPrefix = getLablePrefix("if");

            Label elseLabel = new Label( labelPrefix + ".else");
            Label exitLabel = new Label( labelPrefix + ".exit");
            backend.emitBEQ(ZERO, tmpReg, elseLabel, "IfStmt: If condition is false, jump to else.");
            for(Stmt stmt : node.thenBody) {
                stmt.dispatch(this);
            }
            backend.emitJ(exitLabel, "IfStmt: THEN body done. jump to end.");
            backend.emitLocalLabel(elseLabel, "IfStmt: ELSE");
            for(Stmt stmt : node.elseBody) {
                stmt.dispatch(this);
            }
            backend.emitLocalLabel(exitLabel, "IfStmt: DONE");

            return null;
        }

        @Override
        public Void analyze(WhileStmt node) {
            String labelPrefix = getLablePrefix("while");

            Label beginLabel = new Label(labelPrefix + ".begin");
            Label exitLabel = new Label(labelPrefix + ".exit");

            backend.emitLocalLabel(beginLabel, "WhileStmt: Begin");
            node.condition.dispatch(this);
            RiscVBackend.Register tmpReg = /*borrowOneTmp*/ T6;
            betterBackend.emitPop(tmpReg, "GetCondition");
            backend.emitBEQ(ZERO, tmpReg, exitLabel, "WhileSTmt, IfFalse, exit");

            for(Stmt stmt : node.body) {
                stmt.dispatch(this);
            }
            backend.emitJ(beginLabel, "WhileStmt: GOTO begin");

            backend.emitLocalLabel(exitLabel, "WhileStmt: End");
            return null;
        }

        @Override
        public Void analyze(ReturnStmt node) {
            node.value.dispatch(this);
            betterBackend.emitPop(A0, "ReturnStmt: Get value");
            betterBackend.emitFunctionReturn(this.funcInfo.getFuncName(), "ReturnStmt: GO");
            return null;
        }

        @Override
        public Void analyze(CallExpr node) {
            SymbolInfo called = sym.get(node.function.name);


            if (called instanceof FuncInfo) {
                FuncInfo func = (FuncInfo) called;
                // TODO: Push the FUCKING arg0 (outter function frame ptr)
                if(this.funcInfo != null) {
                    // This MAYBE a fucking nested function call!
                    // FIXME
                }
                for (Expr expr : node.args) {
                    expr.dispatch(this);
                    // WARNING: ALL EXPR SHOULD NEVER EDIT STACK POINTER
                    //          THE EXPR PUSH ITS RETURN VAL TO STACK, AS FUNCTION ARG

                    if(func.getFuncName().equals("print")) {
                        // FUCKING SILLY FUCKING NOOB IDOLT FUNCTION PRINT.
                        if(expr.getInferredType().equals(SymbolType.INT_TYPE) || expr.getInferredType().equals(SymbolType.BOOL_TYPE))
                            betterBackend.emitWrapIntValToIntObject(T6, expr.getInferredType());
                    }
                }

                betterBackend.emitCall(func.getCodeLabel(), null);

                for (Object arg : node.args)
                    betterBackend.emitPop(null, "Pop function arguments.");

                betterBackend.emitPush(A0, "CallExpr: Push function return val");
                return null;
            } else if (called instanceof ClassInfo) {
                // TODO: Fuck the fucking framework.
            }
            return null;
        }

        @Override
        public Void analyze(BooleanLiteral literal) {
            /* Push boolean literal onto stack, incrementing stack pointer*/
            backend.emitLI(A0, (literal.value == true) ? 1 : 0, "Load boolean literal into a0");
            betterBackend.emitPush(A0, "Push literal onto stack");
            return null;
        }

        @Override
        public Void analyze(IntegerLiteral literal) {
            backend.emitLI(A0, literal.value, "Load integer literal into a0");
            betterBackend.emitPush(A0, "Push literal onto stack");
            return null;
        }

        @Override
        public Void analyze(StringLiteral literal) {
            backend.emitLA(A0, constants.getStrConstant(literal.value), "Load string addr into a0");
            betterBackend.emitPush(A0, "Push literal onto stack");
            return null;
        }

        @Override
        public Void analyze(Identifier id) {
            SymbolInfo symbolInfo = sym.get(id.name);
            RiscVBackend.Register tmpReg = T6;

            if (symbolInfo instanceof GlobalVarInfo) {     // global var case
                backend.emitLW(tmpReg, ((GlobalVarInfo) symbolInfo).getLabel(), "Load identifier");
                betterBackend.emitPush(tmpReg, "Ident: GlobalVar");
            } else if (symbolInfo instanceof StackVarInfo){ // local variable
                StackVarInfo stackVarInfo = (StackVarInfo) symbolInfo;
                this.emitLoadNestedParentVar(stackVarInfo, tmpReg, null); // Also work for local variable or param.
                betterBackend.emitPush(tmpReg, "Ident: Arg or stack var");
            }
            return null;
        }

        @Override
        public Void analyze(BinaryExpr binaryExpr) {
            switch (binaryExpr.operator) {
                case "+":
                    if (binaryExpr.left.getInferredType().equals(SymbolType.INT_TYPE)) {
                        binaryExpr.left.dispatch(this);
                        binaryExpr.right.dispatch(this);
                        betterBackend.emitPop(T0, "Load operand 2 into t0");
                        betterBackend.emitPop(T1, "Load operand 1 into t1");
                        
                        backend.emitADD(T0, T0, T1, "Add operands");
                        betterBackend.emitPush(T0, "Push result onto stack");
                        return null;
                    }
                case "-":
                    if (binaryExpr.getInferredType().equals(SymbolType.INT_TYPE)) {    
                        binaryExpr.left.dispatch(this);
                        binaryExpr.right.dispatch(this);
                        betterBackend.emitPop(T0, "Load operand 2 into t0");
                        betterBackend.emitPop(T1, "Load operand 1 into t1");
                        
                        backend.emitSUB(T0, T1, T0, "Subtract operands");
                        betterBackend.emitPush(T0, "Push result onto stack");
                        return null;
                    }
                case "*":
                    if (binaryExpr.getInferredType().equals(SymbolType.INT_TYPE)) {    
                        binaryExpr.left.dispatch(this);
                        binaryExpr.right.dispatch(this);
                        betterBackend.emitPop(T0, "Load operand 2 into t0");
                        betterBackend.emitPop(T1, "Load operand 1 into t1");
                        
                        backend.emitMUL(T0, T1, T0, "Subtract operands");
                        betterBackend.emitPush(T0, "Push result onto stack");
                        return null;
                    }
                case "//":
                    if (binaryExpr.getInferredType().equals(SymbolType.INT_TYPE)) {    
                        binaryExpr.left.dispatch(this);
                        binaryExpr.right.dispatch(this);
                        betterBackend.emitPop(T0, "Load operand 2 into t0");
                        betterBackend.emitPop(T1, "Load operand 1 into t1");

                        backend.emitDIV(T0, T1, T0, "Subtract operands");
                        betterBackend.emitPush(T0, "Push result onto stack");
                        return null;
                    }
                case "%":
                    if (binaryExpr.getInferredType().equals(SymbolType.INT_TYPE)) {    
                        binaryExpr.left.dispatch(this);
                        binaryExpr.right.dispatch(this);
                        betterBackend.emitPop(T0, "Load operand 2 into t0");
                        betterBackend.emitPop(T1, "Load operand 1 into t1");
                        
                        backend.emitREM(T0, T1, T0, "Subtract operands");
                        betterBackend.emitPush(T0, "Push result onto stack");
                        return null;
                    }
                case "==":
                    if (binaryExpr.getInferredType().equals(SymbolType.BOOL_TYPE) ||
                            binaryExpr.getInferredType().equals(SymbolType.INT_TYPE) ) {
                        binaryExpr.left.dispatch(this);
                        binaryExpr.right.dispatch(this);
                        betterBackend.emitPop(T0, "Load operand 2 into t0");
                        betterBackend.emitPop(T1, "Load operand 1 into t1");
                        
                        backend.emitXOR(T0, T0, T1, "== operator on BOOL/INT");
                        backend.emitSEQZ(T0, T0, "Set to 1 if XOR results in 0, they were equal");
                        betterBackend.emitPush(T0, "Push result onto stack");
                        return null;
                    }

                case "!=":
                    if (binaryExpr.getInferredType().equals(SymbolType.BOOL_TYPE) ||
                            binaryExpr.getInferredType().equals(SymbolType.INT_TYPE) ) {
                        binaryExpr.left.dispatch(this);
                        binaryExpr.right.dispatch(this);
                        betterBackend.emitPop(T0, "Load operand 2 into t0");
                        betterBackend.emitPop(T1, "Load operand 1 into t1");
                        
                        backend.emitXOR(T0, T0, T1, "!= operator on BOOL/INT");
                        backend.emitSNEZ(T0, T0, "Set to 1 if XOR was not 0, they were not equal");
                        // Don't think we need to do the SEQZ/SNEZ instructions here because XOR will already result in 1
                        // if they are not equal, which is the desired result.
                        betterBackend.emitPush(T0, "Push result onto stack");
                        return null;
                    }
                case "<=":
                    if (binaryExpr.left.getInferredType().equals(SymbolType.INT_TYPE)) {
                        binaryExpr.left.dispatch(this);
                        binaryExpr.right.dispatch(this);
                        betterBackend.emitPop(T0, "Load operand 2 into t0");
                        betterBackend.emitPop(T1, "Load operand 1 into t1");
                        
                        backend.emitADDI(T1, T1, -1, "decrement LEFT by 1");
                        backend.emitSLT(T1, T1, T0, "LEFT <=? RIGHT");
                        betterBackend.emitPush(T1, "Push result onto stack");
                        return null;
                    }
                case ">=":
                    if (binaryExpr.left.getInferredType().equals(SymbolType.INT_TYPE)) {
                        binaryExpr.left.dispatch(this);
                        binaryExpr.right.dispatch(this);
                        betterBackend.emitPop(T0, "Load operand 2 into t0");
                        betterBackend.emitPop(T1, "Load operand 1 into t1");
                        
                        backend.emitADDI(T0, T0, -1, "decrement RIGHT by 1");
                        backend.emitSLT(T1, T0, T1, "LEFT >=? RIGHT");
                        betterBackend.emitPush(T1, "Push result onto stack");
                        return null;
                    }
                case "<":
                    if (binaryExpr.left.getInferredType().equals(SymbolType.INT_TYPE)) {
                        binaryExpr.left.dispatch(this);
                        binaryExpr.right.dispatch(this);
                        betterBackend.emitPop(T0, "Load operand 2 into t0");
                        betterBackend.emitPop(T1, "Load operand 1 into t1");
                        
                        backend.emitSLT(T1, T1, T0, "LEFT <? RIGHT");
                        betterBackend.emitPush(T1, "Push result onto stack");
                        return null;
                    }
                case ">":
                    if (binaryExpr.left.getInferredType().equals(SymbolType.INT_TYPE)) {
                        binaryExpr.left.dispatch(this);
                        binaryExpr.right.dispatch(this);
                        betterBackend.emitPop(T0, "Load operand 2 into t0");
                        betterBackend.emitPop(T1, "Load operand 1 into t1");
                        
                        backend.emitSLT(T1, T0, T1, "LEFT >? RIGHT");
                        betterBackend.emitPush(T1, "Push result onto stack");
                        return null;
                    }
                case "and":
                    if (binaryExpr.getInferredType().equals(SymbolType.BOOL_TYPE)) {
                        binaryExpr.left.dispatch(this);
                        betterBackend.emitPop(T1, "Load operand 1 into t1");
                        
                        Label skipSecond = generateLocalLabel();
                        backend.emitBEQZ(T1, skipSecond, "If operand 1 is FALSE, don't eval second");
                        binaryExpr.right.dispatch(this);
                        betterBackend.emitPop(T0, "Load operand 2 into t0");
                        backend.emitAND(T1, T1, T0, "AND operands");
                        backend.emitLocalLabel(skipSecond, "After second operand");
                        betterBackend.emitPush(T1, "Push result onto stack");
                        return null;
                    }
                case "or":
                    if (binaryExpr.getInferredType().equals(SymbolType.BOOL_TYPE)) {
                        binaryExpr.left.dispatch(this);
                        betterBackend.emitPop(T0, "Load operand 1 into t0");
                        
                        Label skipSecond = generateLocalLabel();
                        backend.emitBNEZ(T0, skipSecond, "If first is not zero (it's true), skip second");
                        binaryExpr.right.dispatch(this);
                        betterBackend.emitPop(T0, "Load operand 2 into t0");
                        backend.emitOR(T0, T1, T0, "OR operands");
                        backend.emitLocalLabel(skipSecond, "Skip past second operand");
                        betterBackend.emitPush(T0, "Push result onto stack");
                        return null;
                    }
            }

            return null;
        }

        @Override
        public Void analyze(UnaryExpr unaryExpr) {
            switch (unaryExpr.operator) {
                case "-":
                    if (unaryExpr.getInferredType().equals(SymbolType.INT_TYPE)) {
                        unaryExpr.operand.dispatch(this);
                        betterBackend.emitPop(T0, "Load operand into t0");
                        
                        backend.emitSUB(T0, ZERO, T0, "Subtract operands");
                        betterBackend.emitPush(T0, "Push result onto stack");
                        return null;
                    }
                case "not":
                    if (unaryExpr.getInferredType().equals(SymbolType.BOOL_TYPE)) {
                        unaryExpr.operand.dispatch(this);
                        betterBackend.emitPop(T0, "Load operand into t0");
                        
                        backend.emitSEQZ(T0, T0, "NOT operands");
                        betterBackend.emitPush(T0, "Push result onto stack");
                        return null;
                    }
            }
            return null;
        }


        // FIXME: More, of course.

    }

    /**
     * Emits custom code in the CODE segment.
     *
     * This method is called after emitting the top level and the
     * function bodies for each function.
     *
     * You can use this method to emit anything you want outside of the
     * top level or functions, e.g. custom routines that you may want to
     * call from within your code to do common tasks. This is not strictly
     * needed. You might not modify this at all and still complete
     * the assignment.
     *
     * To start you off, here is an implementation of three routines that
     * will be commonly needed from within the code you will generate
     * for statements.
     *
     * The routines are error handlers for operations on None, index out
     * of bounds, and division by zero. They never return to their caller.
     * Just jump to one of these routines to throw an error and
     * exit the program. For example, to throw an OOB error:
     *   backend.emitJ(errorOob, "Go to out-of-bounds error and abort");
     *
     */
    protected void emitCustomCode() {
//        backend.emitGlobalLabel(new Label("int-alloc"));
//        backend.emitLA(A0, intClass.getPrototypeLabel(), "Get int prototype address");
//        backend.emitJAL(objectAllocLabel, "Allocate int");
//
//        FuncInfo allocInt = makeFuncInfo("int-alloc", 0, SymbolType.OBJECT_TYPE,
//                globalSymbols, null, this::emitStdFunc);
//        allocInt.addParam(makeStackVarInfo("arg", SymbolType.OBJECT_TYPE,
//                null, allocInt));
//        functions.add(allocInt);
//        System.out.println(allocInt.getBaseName());
//        globalSymbols.put(allocInt.getBaseName(), allocInt);
        emitErrorFunc(errorNone, "Operation on None");
        emitErrorFunc(errorDiv, "Divison by zero");
        emitErrorFunc(errorOob, "Index out of bounds");
    }

    /** Emit an error routine labeled ERRLABEL that aborts with message MSG. */
    private void emitErrorFunc(Label errLabel, String msg) {
        backend.emitGlobalLabel(errLabel);
        backend.emitLI(A0, ERROR_NONE, "Exit code for: " + msg);
        backend.emitLA(A1, constants.getStrConstant(msg),
                       "Load error message as str");
        backend.emitADDI(A1, A1, getAttrOffset(strClass, "__str__"),
                         "Load address of attribute __str__");
        backend.emitJ(abortLabel, "Abort");
    }
}
